<%@ webhandler language="c#" class="ContextViewer" debug="true" %>

using System;
using System.Web;
using System.Reflection;
using System.Runtime.Remoting.Messaging;
using System.Threading;
using System.Diagnostics;

public class ContextViewer : IHttpAsyncHandler
{
    static DevelopMentor.ThreadPool threadPool;
    
    static ContextViewer()
    {
        // Issue: the first request to hit this handler boostraps the
        //        new appdomain & authentication scheme; the result of which
        //        is then used to initialize the thread pool.  This means the
        //        thread identity of the initial threads (2 in the code below)
        //        is equivalent to the thread identity used for this first request.
        //        Similarly, any dynamically added threads will inherit the
        //        identity of the thread that happens to trigger the new thread
        //        being added to the pool.
        //
        threadPool = new DevelopMentor.ThreadPool(2, 5, "Test Pool");
        threadPool.PropogateCallContext = true;
        threadPool.PropogateThreadPrincipal = true;
        threadPool.PropogateHttpContext = true;
        threadPool.Start();
    }
    
    // IHttpHandler
    //
    public bool IsReusable { get { return(true); } }
    
    public void ProcessRequest( HttpContext ctx )
    {
        ctx.Response.Output.Write("*** Should not see this ***<br/>");
    }
    
    // IHttpAsyncHandler
    //
    public IAsyncResult BeginProcessRequest( HttpContext ctx, AsyncCallback cb, object extraData )
    {
        BreakIfRequested();

        ctx.Response.Buffer = false;
                
        ctx.Response.Output.Write( "IHttpAsyncHandler.BeginProcessRequest called on thread {0}<br/>",
                                   AppDomain.GetCurrentThreadId() );
        ctx.Response.Output.Write("HttpContext.Current = {0}<br/>", (HttpContext.Current == null ? "n/a" : HttpContext.Current.ToString()));

        AsyncRequestState reqState = new AsyncRequestState(ctx, cb, extraData);
        threadPool.PostRequest(new DevelopMentor.WorkRequestDelegate(ProcessRequest), reqState);
        return(reqState);
    }
    
    public void EndProcessRequest( IAsyncResult ar )
    {
        AsyncRequestState reqState = (AsyncRequestState)ar;

        reqState.ctx.Response.Output.Write( "IHttpAsyncHandler.EndProcessRequest called on thread {0}<br/>",
                                            AppDomain.GetCurrentThreadId() );
    }
    
    void ProcessRequest( object state, DateTime requestTime )
    {
        AsyncRequestState reqState = (AsyncRequestState)state;
        HttpContext ctx = reqState.ctx;
        
        ctx.Response.Output.Write( "Handler.ProcessRequest called on thread {0}<br/>",
                                   AppDomain.GetCurrentThreadId() );
        ctx.Response.Output.Write("HttpContext.Current = {0}<br/>", (HttpContext.Current == null ? "n/a" : HttpContext.Current.ToString()));
        ctx.Response.Output.Write( "IHttpHandler.ProcessRequest HttpContext principal: {0}<br/>",
                                   ctx.User.Identity.Name );
        ctx.Response.Output.Write( "IHttpHandler.ProcessRequest thread principal: {0}<br/>",
                                   Thread.CurrentPrincipal.Identity.Name );
        ctx.Response.Output.Write("Handler.ProcessRequest started sleeping at {0}<br/>", DateTime.Now);
        
        for( int n = 0; n < 5; n++ )
        {
            Thread.Sleep(1000);
            ctx.Response.Output.Write("Handler.ProcessRequest still sleeping at {0}<br/>", DateTime.Now);
        }
        
        ctx.Response.Output.Write("Handler.ProcessRequest finished sleeping at {0}<br/>", DateTime.Now);

        // This triggers the asp.net plumbing to call our
        // IHttpAsyncHandler.EndProcessRequest method.
        //
        reqState.CompleteRequest();
        reqState.cb(reqState);
    }

    
    class AsyncRequestState : IAsyncResult
    {
        public AsyncRequestState( HttpContext ctx, AsyncCallback cb, object extraData )
        {
            this.ctx = ctx;
            this.cb = cb;
            this.extraData = extraData;
        }
        
        internal HttpContext ctx;
        internal AsyncCallback cb;
        internal object extraData;
        private bool isCompleted = false;
        private ManualResetEvent callCompleteEvent = null;
        
        internal void CompleteRequest()
        {
            isCompleted = true;
            
            lock( this )
            {
                if( callCompleteEvent != null )
                {
                    callCompleteEvent.Set();
                }
            }
        }
        
        // IAsyncResult
        //
        public object AsyncState { get { return(extraData); } }
        public bool CompletedSynchronously { get { return(false); } }
        public bool IsCompleted { get { return(isCompleted); } }
        
        public WaitHandle AsyncWaitHandle
        {
            get
            {
                lock( this )
                {
                    if( callCompleteEvent == null )
                    {
                        callCompleteEvent = new ManualResetEvent(false);
                    }
                    
                    return(callCompleteEvent);
                }
            }
        }
    }
    
    void BreakIfRequested()
    {
        string breakArg = HttpContext.Current.Request.QueryString["break"];
        
        if( (breakArg != null) && ((breakArg == "yes") || (breakArg == "true")) )
        {
            if( Debugger.IsAttached )
            {
                Debugger.Break();
            }
            else
            {
                Debugger.Launch();
            }
        }
    }
}
